
/*
// author : Sven Goethel - Stapenhorststrasse 35a - 33615 Bielefeld
// A polymorph list
// index-range: [0-n[ , where n is the number of elements !
*/

/*
	#define __LISTE_PARANOIA__
	#define __VERBOSE_LISTE_PARANOIA__
*/

#include <string.h>
#include <stdlib.h>

#ifdef _C166_
  #define _ERROR_OUT_ printf(
#else
  #define _ERROR_OUT_ fprintf(stderr,
#endif

#include "rtti.h"

typedef struct ListElem
{
  rtti        * element;  /* the Element */
  void        * pred;     /* the predecessor */
  void        * succ;     /* the successor */
  void        * lh;       /* the owner list, need to set the acct and first */
                          /* pointer and to recalc the memory-size */
} ListElem;

typedef struct CListe
{
  rtti         rtype;           /* rtti for the CListe */
  ListElem     * first;         /* the first-element */
  ListElem     * acct;          /* the actual-element  */
  int           acct_index;    /* the index of the actual-element */
  int           number;        /* the number of elements */
  unsigned long memory;        /* the size of the whole allocated memory  */
                               /* for the elements  */
                               /* Sum(sizeof(ListElem)+ListElem[i].sz) */
} CListe;

#define PLISTELEM(a) ((ListElem *)((a)))
typedef CListe * HCListe;


#ifdef __LISTE_PARANOIA__
  void CListeParanoiaCheck(void);
  static HCListe __lh;
#endif

#define __CLISTE__

#include "cliste.h"

/*
 *          L i s t E l e m  -- M e t h o d s
 *
*/

/* fully recalculates the CListe */
static void ListElem_Destroy(ListElem *le)
{
  if(le==0) INT_ERR(__LINE__);
  if(le->lh==0) INT_ERR(__LINE__);
  if(le->element==0) INT_ERR(__LINE__);
  if( ((HCListe)(le->lh))->memory < le->element->sz+sizeof(*le) )
    INT_ERR(__LINE__);
  ((HCListe)(le->lh))->memory-=le->element->sz+sizeof(*le);

  DestroyRtti(le->element);
  free(le);
}

/* fully recalculates the CListe */
static void ListElem_Copy(ListElem *DestLe, ListElem *SrcLe)
{
  if(SrcLe==0) INT_ERR(__LINE__);
  if(DestLe==0) INT_ERR(__LINE__);
  if(DestLe->lh==0) INT_ERR(__LINE__);
  if(SrcLe->lh==0) INT_ERR(__LINE__);
  if(DestLe->element!=0) INT_ERR(__LINE__);
  if(SrcLe->element==0) INT_ERR(__LINE__);

  DestLe->element=CopyCreateRtti(SrcLe->element);
  ((HCListe)(DestLe->lh))->memory+=DestLe->element->sz;
}

/* fully recalculates the CListes memory */
static ListElem *ListElem_CreateBlank(CListe *lh)
{
  ListElem *res=NULL;
  if(lh==0) INT_ERR(__LINE__);
  res=(ListElem *)calloc(1, sizeof(*res));
  if(res==0) INT_ERR(__LINE__);
	#ifdef __LISTE_PARANOIA__
  	if(sizeof(rtti *)!=sizeof(void *))
    {
      ErrorOut("Watch sizeof");
      ErrorOut("sizeof(rtti *) = %u", sizeof(rtti *));
      ErrorOut("sizeof(void *) = %u", sizeof(void *));
      INT_ERR(__LINE__);
    }
		if(sizeof(*res)!=sizeof(ListElem))
    {
      ErrorOut("Watch sizeof");
      ErrorOut("sizeof(*res) = %u", sizeof(*res));
      ErrorOut("sizeof(ListElem) = %u", sizeof(ListElem));
      INT_ERR(__LINE__);
    }
		if(res->element!=0) { ErrorOut("Watch calloc"); INT_ERR(__LINE__); }
  #endif
  res->lh=(void *)lh;
  ((HCListe)(res->lh))->memory+=sizeof(*res);
  return res;
}

static ListElem *ListElem_Create(CListe *lh, const rtti *element)
{
  ListElem *res=NULL;
  if(lh==0) INT_ERR(__LINE__);
  if(element==0) INT_ERR(__LINE__);

  res=ListElem_CreateBlank(lh);
  res->element=CopyCreateRtti(element);
  ((HCListe)(res->lh))->memory+=res->element->sz;
  return res;
}

/* fully recalculates the CListes memory */
static ListElem *ListElem_CopyCreate(ListElem *SrcLe, CListe *lh)
{
  ListElem *res=ListElem_CreateBlank(lh);
  ListElem_Copy(res, SrcLe);
  return res;
}

/*
 *          L i s t H e a d e r  -- M e t h o d s
 *
 *          the memory-member is totally controled by the
 *          allocation and deallocation functions of the
 *          ListElem-Methods !!!!!!!!!!!!!!!!!!!!!!!!!!!!
*/

static rtti * RttiCListe = NULL;

static void CListe_Reset(void)
{
  RttiCListe=NULL;
}

const rtti * const GetRttiCListe(void)
{ if(RttiCListe==NULL) INT_ERR(__LINE__); return RttiCListe; }

const rtti * const CListeRegisterRtti       (void)
{
  if(RttiCListe == NULL)
  {
    RttiCListe = NewRttiType(sizeof(CListe));
    RegisterRttiFct(RttiCListe,
                    (T_FCmpRtti) CListe_IsEqual,
                    (T_FCreateRtti) CListe_Create,
                    (T_FCopyCreateRtti) CListe_CopyCreate,
                    (T_FDestroyRtti) CListe_Destroy,
                    (T_FResetRtti) CListe_Reset
                   );
  }
  return GetRttiCListe();
}


static void CListe_Anfang(HCListe lh)
{
  lh->acct=lh->first;
  lh->acct_index=0;
}

static void CListe_GoFwd(HCListe lh)
{
  if(lh == 0) INT_ERR(__LINE__);
  if(lh->acct == 0) INT_ERR(__LINE__);
  if(lh->acct_index >= lh->number) INT_ERR(__LINE__);
  if(lh->acct->succ==0) INT_ERR(__LINE__);

  lh->acct_index++;
  lh->acct = lh->acct->succ;

  if(lh->acct == 0) INT_ERR(__LINE__);
}

static void CListe_GoBwd(HCListe lh)
{
  if(lh == 0) INT_ERR(__LINE__);
  if(lh->acct == 0) INT_ERR(__LINE__);
  if(lh->acct_index == 0) INT_ERR(__LINE__);
  if(lh->acct->pred==0) INT_ERR(__LINE__);

  lh->acct_index--;
  lh->acct = lh->acct->pred;

  if(lh->acct == 0) INT_ERR(__LINE__);
}

static void CListe_Copy(HCListe DestLh, HCListe SrcLh)
{
  int number=0;
  ListElem *mZgr   = NULL;
  ListElem *neuZgr = NULL;

  if(DestLh==0) INT_ERR(__LINE__);
  if(DestLh->number>0) INT_ERR(__LINE__);
  if(DestLh->first!=0) INT_ERR(__LINE__);
  if(SrcLh==0) INT_ERR(__LINE__);
  if ((DestLh->number = SrcLh->number) == 0)
  {
    CListe_Anfang(DestLh);
    return;
  }

  DestLh->acct_index = SrcLh->acct_index;
  mZgr   = SrcLh->first;
  neuZgr = DestLh->acct= DestLh->first =
           (void *)ListElem_CopyCreate(mZgr, DestLh);
  number++;

  while((mZgr = PLISTELEM(mZgr->succ)) != 0)
  {
    number++;
    neuZgr->succ = (void *) ListElem_CopyCreate(mZgr, DestLh);
    PLISTELEM(neuZgr->succ)->pred = (void *)neuZgr;
    neuZgr = PLISTELEM(neuZgr->succ);
    if (mZgr == SrcLh->acct) DestLh->acct = neuZgr;
  }
  if(DestLh->number!=number) INT_ERR(__LINE__);
  #ifdef __LISTE_PARANOIA__
	  __lh=DestLh;
    CListeParanoiaCheck();
	  __lh=SrcLh;
    CListeParanoiaCheck();
  #endif
}

int     CListe_Laenge          (const HCListe lh)
{
  if(lh == 0) INT_ERR(__LINE__);
  return lh->number;
}

void    CListe_Destroy        (HCListe lh)
{
  if(lh == 0) INT_ERR(__LINE__);
  CListe_Clear(lh);
  free(lh);
}

void    CListe_Clear          (HCListe lh)
{
  if(lh == 0) INT_ERR(__LINE__);

  while (lh->first)
  {
    ListElem *next=PLISTELEM(lh->first->succ);
    ListElem_Destroy(lh->first);
    lh->first=next;
    lh->number--;
  }
  CListe_Anfang(lh);

  if(lh->memory!=0) INT_ERR(__LINE__);
}

rtti * CListe_Create(void)
{
  HCListe res=NULL;
  CListeRegisterRtti();
  res=(CListe *)calloc(1,sizeof(*res));
  if(!res) INT_ERR(__LINE__);
	#ifdef __LISTE_PARANOIA__
		if(res->number!=0) { ErrorOut("Watch calloc"); INT_ERR(__LINE__); }
		if(res->memory!=0) { ErrorOut("Watch calloc"); INT_ERR(__LINE__); }
		if(res->acct_index!=0) { ErrorOut("Watch calloc"); INT_ERR(__LINE__); } 
  #endif
  res->rtype.type=RttiCListe->type;
  res->rtype.sz=RttiCListe->sz;
  return C_RTTI(res);
}

rtti * CListe_CopyCreate(const HCListe lh)
{
  HCListe res=NULL;
  if(lh == 0) INT_ERR(__LINE__);
  res=(HCListe)(CListe_Create());
  CListe_Copy(res, lh);
  res->rtype.type=RttiCListe->type;
  res->rtype.sz=RttiCListe->sz;
  return C_RTTI(res);
}

int CListe_IsEqual(const HCListe LhA, const HCListe LhB)
{
  int i;
  if(LhA == 0) INT_ERR(__LINE__);
  if(LhB == 0) INT_ERR(__LINE__);
  if(LhA->number!=LhB->number) return 0;
  for(i=0; i<LhA->number &&
           CmpRtti(CListe_GetElement(LhA,i), CListe_GetElement(LhB,i));
      i++);
  return i == LhA->number;
}

int CListe_FuegeEin(HCListe lh, const rtti *elem, int i)
{
  rtti_result rr;
  int oldInd;
  int k=0;
  ListElem *neu = NULL;
  if(lh == 0) INT_ERR(__LINE__);
  if(elem == 0) INT_ERR(__LINE__);
  oldInd = lh->acct_index;

	if((rr=CheckRtti(C_RTTI(elem)))!=RTTI_OK)
	{
	  ErrorOut("elem check failed");
	  ErrorOut(GetRttiErrorStr(rr));
		INT_ERR(__LINE__);
  }
	if((rr=CheckRtti(C_RTTI(lh)))!=RTTI_OK)
	{
	  ErrorOut("lh check failed");
	  ErrorOut(GetRttiErrorStr(rr));
		INT_ERR(__LINE__);
  }
  if (i < 0 || i > lh->number)
  {
    _ERROR_OUT_ "\n Falscher Index.");
    _ERROR_OUT_ "\n Es wurde versucht ein Element an %d.Stelle einer Liste mit %d Elementen einzufuegen.\n",
     i, lh->number);
    INT_ERR(__LINE__);
  }
  neu = ListElem_Create(lh, elem);
  if (neu == 0)
  {
    _ERROR_OUT_ "\n Speicherfehler.");
    _ERROR_OUT_ "\n Es wurde versucht ein Listen-Element fuer eine Liste mit %d Elementen zu erzeugen.\n",
     lh->number);
    INT_ERR(__LINE__);
  }

	if(lh->number==0 && lh->first!=NULL
		 ||
		 lh->number!=0 && lh->first==NULL)
	{
		ErrorOut("first : %p", lh->first);
		ErrorOut("acct : %p", lh->acct);
		ErrorOut("acct_index : %i", lh->acct_index);
		ErrorOut("number : %i", lh->number);
		ErrorOut("memory : %lu", lh->memory);
		if(lh->number==0 && lh->first!=NULL)
			INT_ERR(__LINE__);
		if(lh->number!=0 && lh->first==NULL)
			INT_ERR(__LINE__);
	}
  if (i == 0)
  {
    neu->succ = lh->first;
    neu->pred = 0;
    if (lh->first) lh->first->pred = (void *)neu;
    lh->first = neu;
    CListe_Anfang(lh);
  }
  else
  {
    i--; // Das Element 'ein' vor dem Ziel anfahren !
    if (i < oldInd)
      for (k = oldInd; k > i; k--) CListe_GoBwd(lh);
    else if (i > oldInd)
      for (k = oldInd; k < i; k++) CListe_GoFwd(lh);
    neu->succ = lh->acct->succ;
    neu->pred = (void *)(lh->acct);
    if (lh->acct->succ) PLISTELEM(lh->acct->succ)->pred = (void *)neu;
    lh->acct->succ = (void *)neu;
  }

  lh->number+=1;

  #ifdef __LISTE_PARANOIA__
	  __lh=lh;
    CListeParanoiaCheck();
  #endif
  return 1;
}

void CListe_Loesche(HCListe lh, int i)
{
  int oldInd;
  int k=0;
  void *succ=NULL;
  void *pred =NULL;

  if(lh == 0) INT_ERR(__LINE__);
  oldInd = lh->acct_index;

  if (i < 0 || i >= lh->number)
  {
    _ERROR_OUT_ "\n Falscher Index.");
    _ERROR_OUT_ "\n Es wurde auf ein Element an %d.Stelle einer Liste mit %d Elementen zu loeschen.\n",
     i, lh->number);
    INT_ERR(__LINE__);
  }

  if (i < oldInd)
  {
    for (k = oldInd; k > i; k--)
      CListe_GoBwd(lh);
  }
  else if (i > oldInd)
  {
     for(k = oldInd; k < i; k++)
      CListe_GoFwd(lh);
  }
  succ=lh->acct->succ;
  pred =lh->acct->pred;

  if (lh->acct == lh->first)
    lh->first = succ;
  else
    PLISTELEM(pred)->succ = succ;
  if (succ) PLISTELEM(succ)->pred = pred;
  ListElem_Destroy(lh->acct);
  lh->number--;
  if (succ)
    lh->acct=succ;
  else if (pred)
  {
    lh->acct=pred;
    lh->acct_index--;
  }
  else CListe_Anfang(lh);
  if ( lh->acct_index < 0 ) INT_ERR(__LINE__);
  if (!( ( lh->number==0 && lh->acct_index==0 && lh->acct==NULL &&
             lh->first==NULL
           )
           ||
           ( lh->number && lh->acct!=NULL && lh->first!=NULL )
         ) ) INT_ERR(__LINE__);
  #ifdef __LISTE_PARANOIA__
	  __lh=lh;
    CListeParanoiaCheck();
  #endif
}

rtti * CListe_GetElement(const HCListe lh, int i)
{
  int oldInd=0;
  int k=0;

  if(lh == 0) INT_ERR(__LINE__);
  oldInd = lh->acct_index;

  if (i < 0 || i >= lh->number)
  {
    _ERROR_OUT_ "\n Falscher Index.");
    _ERROR_OUT_ "\n Es wurde auf ein Element an %d.Stelle einer Liste mit %d Elementen zugegriffen.\n",
       i, lh->number);
    INT_ERR(__LINE__);
  }

  if (i < oldInd)
  {
    for (k = oldInd; k > i; k--)
      CListe_GoBwd(lh);
  }
  else if (i > oldInd)
  {
    for(k = oldInd; k < i; k++)
      CListe_GoFwd(lh);
  }
  #ifdef __LISTE_PARANOIA__
	  __lh=lh;
    CListeParanoiaCheck();
  #endif
  return lh->acct->element;
}

int CListe_IstElement(const HCListe lh, const rtti * element, int i)
{
  if(lh == 0) INT_ERR(__LINE__);
  if(element == 0) INT_ERR(__LINE__);

  #ifdef __LISTE_PARANOIA__
	  __lh=lh;
    CListeParanoiaCheck();
  #endif

  for ( ; i < lh->number; i++)
  {
    if ( CmpRtti(CListe_GetElement(lh, i), element) ) return i;
  }
  return -1;
}

int CListe_Referenz2Index(const HCListe lh, const rtti * element)
{
  int i;
  if(lh == 0) INT_ERR(__LINE__);
  if(element == 0) INT_ERR(__LINE__);

  #ifdef __LISTE_PARANOIA__
	  __lh=lh;
    CListeParanoiaCheck();
  #endif

  i=lh->number-1;
  while(i>=0 && element!=CListe_GetElement(lh, i)) i--;
  return i;
}

unsigned long  CListe_GetMemoryUsage(const HCListe lh)
{
  return sizeof(*lh)+(lh->memory);   /* the size of the whole allocated memory  */
}

#ifdef __LISTE_PARANOIA__
    static ListElem *pi;
		static int i;
  void CListeParanoiaCheck(void)
  {
    pi=NULL;
    i=0;
    if(__lh == 0) INT_ERR(__LINE__);

    pi=__lh->first;
    while ( pi )
    {
      i++;
      if(pi->pred!=NULL)
      {
        if(PLISTELEM(pi->pred)->succ!=pi)
				{
					#ifdef __VERBOSE_LISTE_PARANOIA__
						ErrorOut("ListElement #%d\n\r", i);
						ErrorOut(" pi             : %p\n\r", pi);
						ErrorOut(" pi->pred       : %p\n\r", pi->pred);
						ErrorOut(" pi->pred->succ : %p\n\r", PLISTELEM(pi->pred)->succ);
					#endif
          INT_ERR(__LINE__);
        }
      }
      if(pi->succ!=NULL)
      {
        if(PLISTELEM(pi->succ)->pred!=pi)
				{
					#ifdef __VERBOSE_LISTE_PARANOIA__
						ErrorOut("ListElement #%d\n\r", i);
						ErrorOut(" pi             : %p\n\r", pi);
						ErrorOut(" pi->succ       : %p\n\r", pi->succ);
						ErrorOut(" pi->succ->pred : %p\n\r", PLISTELEM(pi->succ)->pred);
					#endif
          INT_ERR(__LINE__);
        }
      }
      pi=pi->succ;
    }
    if(i!=__lh->number)
		{
		  ErrorOut("counted number : %d, saved number : %d", i, __lh->number);
      INT_ERR(__LINE__);
    }
  }
#endif

