/*************************************************************************/
/*
 *  Datei           : metageom.cc
 *  Inhalt          : Definitionen fuer Klassen :
 *  Programmierer   : Sven Goethel
 *            Stapenhorststr.35a
 *            33615 Bielefeld
 *            0521/139228
 *
 *  $Log: metageom.cc $
 *  Revision 1.2  1995/04/07 16:54:54  Sven_Goethel_DFP
 *  *** empty log message ***
 *  Revision 1.1  1995/04/04 10:37:30  Sven_Goethel
 *  Initial revision
 */
/*************************************************************************/

#ifdef _Windows
  #include <afx.h>
#else
  #define TRACE(a)
#endif  
# include "metageom.h"

/*------------------------------------------------------------------------------

  ElementFunktion der Klasse MetaGeometry
  --------------------------------------------
  t_real ::AddWinkel(const t_real a, const t_real b)

  - Beschreibung
  --------------
  Addiert Winkel a und Winkel b

  - Parameter
  -----------
  E: a,b  - die Winkel-Summanden


  - allgemeine Ein- und Ausgabeeinheiten
  --------------------------------------
  keine

  - Returncode
  ------------
  0<=Winkel<=360 in Grad !!!!

  - Historie
  ----------
  29.6.1994 Sven Goethel

------------------------------------------------------------------------------*/
void MGeom_AddWinkel(t_real *result, const t_real *a, const t_real *b)
{
  rtti *temp;
	temp=GetRttiReal();
  result->rtype=*temp;
  result->r=a->r+b->r;
  if(result->r>360.0)
    result->r-=360;
}

/*------------------------------------------------------------------------------

  ElementFunktion der Klasse MetaGeometry
  --------------------------------------------
  t_real ::SubWinkel(const t_real a, const t_real b)

  - Beschreibung
  --------------
  Subtrahiert Winkel b von Winkel a  => a-b

  - Parameter
  -----------
  E: a,b  - die Winkel-Summanden


  - allgemeine Ein- und Ausgabeeinheiten
  --------------------------------------
  keine

  - Returncode
  ------------
  0<=Winkel<=360 in Grad !!!! ( a-b )

  - Historie
  ----------
  29.6.1994 Sven Goethel

------------------------------------------------------------------------------*/
void MGeom_SubWinkel(t_real *result, const t_real *a, const t_real *b)
{
  rtti *temp;
	temp=GetRttiReal();
  result->rtype=*temp;
  result->r=a->r-b->r;
  if(result->r<0)
    result->r+=360;
}

/*------------------------------------------------------------------------------

  ElementFunktion der Klasse MetaGeometry
  --------------------------------------------
	      t_real VektorBetrag(const t_coord *P0, const t_coord *PM,
		  const t_coord *P1, t_rotation rot)

  - Beschreibung
  --------------
  Berechnet den Betrag des Vektors von P1 nach P2.

  - Parameter
  -----------
  E: P1, P2 Die beiden t_coorde.

  - allgemeine Ein- und Ausgabeeinheiten
  --------------------------------------
  keine

  - Returncode
  ------------
  der Betrag

  - Historie
  ----------
  16.6.1994 Sven Goethel

------------------------------------------------------------------------------*/
void MGeom_KreisLaenge(t_real *result, const t_coord *P0, const t_coord *PM,
		  const t_coord *P1, t_rotation rot)
{
  t_real Radius, Gamma;
	rtti *temp;
  MGeom_VektorBetrag(&Radius, PM,P0);
  // ... Da dieser innere Winkel genau anders Interpretiert wird,
  // wurde Start und Endpunkt umgedreht !!!
  MGeom_GetWinkel12PPPR ( &Gamma, P1, PM, P0, rot );
  temp=GetRttiReal();
  result->rtype=*temp;
  result->r=Gamma.r*Radius.r*M_PI/180;
}

void MGeom_VektorBetrag(t_real *result, const t_coord *P1, const t_coord *P2)
{
  rtti *temp;
  temp=GetRttiReal();
  result->rtype=*temp;
  result->r=
	 sqrt
	 (
	   fabs(
	     ( P1->x.r-P2->x.r ) *  ( P1->x.r-P2->x.r )
	     +  ( P1->y.r-P2->y.r ) *  ( P1->y.r-P2->y.r )
	     +  ( P1->z.r-P2->z.r ) *  ( P1->z.r-P2->z.r )
	    )
	 ) ;
}

/*------------------------------------------------------------------------------

  ElementFunktion der Klasse Geometry
  --------------------------------------------
  double ::GetWinkelAbsolut(const FxmObject *FObj, Position Pos)

  - Beschreibung
  --------------
  Berechnet den Winkel Absolut zur Positiven X-Achse.


  - Parameter
  -----------
  E: FObj   Die Kontur-Objekte
  E: Pos      Anfang oder Ende

  - allgemeine Ein- und Ausgabeeinheiten
  --------------------------------------
  keine

  - Returncode
  ------------
  den Winkel (0-90(I.),90-180(II.),180-270(III.),270-360(IV.))
  zur X-Achse des I.Quadranten in GRAD....

  - Historie
  ----------
  29.6.1994 Sven Goethel

------------------------------------------------------------------------------*/
void MGeom_GetWinkelAbsolutPPR(t_real *result, const t_coord *Pp, const t_coord *PM,
				      t_rotation Rot)
{
  rtti *temp;
  t_real t;
  temp=GetRttiReal();
  result->rtype=*temp;
  t.rtype=*temp;
  t.r=90;
  // BOGEN, Winkel der Tangente des Bogenendpunktes und der X-Achse bilden.

  // PM - Mittelpunkt
  // Pp - Winkelbildende Punkt

  // Winkel Beta zwischen positiver X-Achse (I.Quadrant)  und PM->Pp,  !
  MGeom_GetWinkelAbsolutPP(result, PM, Pp);
  if(Rot==CCW) MGeom_AddWinkel(result, result, &t);
  else MGeom_SubWinkel(result, result, &t);
}

/*------------------------------------------------------------------------------

  ElementFunktion der Klasse MetaGeometry
  --------------------------------------------
  t_real ::GetWinkelAbsolut(const t_coord *P0, const t_coord *P1)

  - Beschreibung
  --------------
  Berechnet den Winkel Absolut zur Positiven X-Achse.


  - Parameter
  -----------
  E: P0, P1 t_coorde der Linie P0->P1, dessne Winkel berechnet werden soll

  - allgemeine Ein- und Ausgabeeinheiten
  --------------------------------------
  keine

  - Returncode
  ------------
  den Winkel (0-90(I.),90-180(II.),180-270(III.),270-360(IV.))
  zur X-Achse des I.Quadranten in GRAD....

  - Historie
  ----------
  29.6.1994 Sven Goethel

------------------------------------------------------------------------------*/
void MGeom_GetWinkelAbsolutPP(t_real *result, const t_coord *P0, const t_coord *P1)
{
  t_real nn;
  rtti *temp;
	temp=GetRttiReal();
  result->rtype=*temp;

  result->r=0;
  // Winkelberechnung .... 0-90 Grad !
  MGeom_VektorBetrag(&nn, P0, P1);
  if(nn.r == 0.0 ) return;
  result->r=asin(fabs(P1->y.r-P0->y.r)/nn.r);
  // vom Bogenmass zum Gradmass
  result->r=result->r*180/M_PI;

  // Fallunterscheidung; in welchem Quadranten liegt der Winkel .....
  // P1[0]-P0[0] ... Cosinus Vorzeichen (P1.x-P0.x) !
  // P1[1]-P0[1] ... Sinus Vorzeichen   (P1.y-P0.y) !
  if ( P1->x.r-P0->x.r<=0 && P1->y.r-P0->y.r>=0 ) result->r=180-result->r;
  else if ( P1->x.r-P0->x.r<=0 && P1->y.r-P0->y.r<=0 ) result->r=180+result->r;
  else if ( P1->x.r-P0->x.r>=0 && P1->y.r-P0->y.r<=0 ) result->r=360-result->r;
  // else if ( P1->x-P0->x>0 ** P1->y-P0->y>0 ) I.Quadrant keine Aenderung !

}


// Berechne den eingeschlossenen Winkel der Strecken P0->P1 und P1->P2 !!!
// Es wird hier der Cosinussatz verwendet !!
// Falls kein Ergebniss moeglich ist, wird ein Wert < 0.0 returniert !!!
void MGeom_GetWinkel12 ( t_real *result, const t_coord *P0, const t_coord *P1,
				   const t_coord *P2
				 )
{
  t_coord s1, s2, s3;
  double a,b,c;
	t_real help;
  rtti *temp;
	temp=GetRttiReal();
  result->rtype=*temp;
  result->r=0;
  Coord_Sub(&s1,P0,P2);
  Coord_Sub(&s2,P0,P1);
  Coord_Sub(&s3,P1,P2);
  Coord_Length(&help,&s1);
  a=help.r;
  Coord_Length(&help,&s2);
	b=help.r;
  Coord_Length(&help,&s3);
	c=help.r;
  if(b<=EPSILON || c<=EPSILON)
  {
    result->r=-1.0;
  }
  if(b*c==0.0) INT_ERR(__LINE__);

  result->r = acos ( (a*a-b*b-c*c)/(-2*b*c) ) * 180.0 / M_PI;
}

/*------------------------------------------------------------------------------

  ElementFunktion der Klasse MetaGeometry
  --------------------------------------------
  t_real ::GetWinkel12 ( const t_coord P0, const t_coord P1,
	       const t_coord P2
	     );

  - Beschreibung
  --------------
  Berechnet den Winkel zwischen den Linien P0-P1 und P1-P2 !!
  Ist der Endpunkt der Linie ConA nicht gleich Anfangspunkt der Linie
  von ConB, so wird eine negative Zahl ausgegeben !!!


  - Parameter
  -----------
  E: P0, P1 - Die eine Linie
  E: P1, P2 - Die andere Linie

  - allgemeine Ein- und Ausgabeeinheiten
  --------------------------------------
  keine

  - Returncode
  ------------
  den Winkel (0-90(I.),90-180(II.),180-270(III.),270-360(IV.))
  zur X-Achse des I.Quadranten in GRAD....

  - Historie
  ----------
  29.6.1994 Sven Goethel

------------------------------------------------------------------------------*/
void MGeom_GetWinkel12PPPR ( t_real *result, const t_coord *P0, const t_coord *P1,
			       const t_coord *P2, t_rotation rot
			     )
{
  t_real Alpha, Beta;
  MGeom_GetWinkelAbsolutPP(&Alpha, P0, P1);
  MGeom_GetWinkelAbsolutPP(&Beta, P1, P2);
  MGeom_GetWinkel12ABR(result, &Alpha, &Beta, rot);
}

void MGeom_GetWinkel12ABR ( t_real *result, t_real *Alpha, t_real *Beta,
				   t_rotation rot
				 )
{
  t_real a, b;
  t_real t;
  rtti *temp;
	temp=GetRttiReal();
  result->rtype=*temp;

  result->r=0;
  a.r=Alpha->r;
  b.r=Beta->r;

  // Alpha Umdrehen
  t.r=180.0;
  MGeom_AddWinkel(&a, &a,&t);

  // Innerer Winkel Zwischen Alpha und Beta Linksherum berechnen !!!!
  if(b.r>a.r) b.r=b.r-360; // VORZEICHEN WICHTIG
  MGeom_SubWinkel(&a, &a,&b);
  t.r=360.0;
  if(rot==CW) MGeom_SubWinkel(&a, &t, &a);
  result->r=a.r;
}


